
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>       /* for cos(), sin(), and sqrt() */
#include <time.h>
#include <GL/glut.h>

#ifdef WIN32
#include <windows.h>
#endif

float mpsPointSize = 1; // 0 = default OpenGL value
float mpsMaxPointSize;
//int mpsLineSize; // 0 = default OpenGL value

static int mpsLength;
typedef float mpsVectorsType [3];
typedef float mpsDsType [3];
typedef float mpsColorsType [4];

static int *mpsIndices;

static mpsVectorsType *mpsVectors;
static mpsDsType *mpsDs;
static mpsColorsType *mpsColors;

static mpsLockArraysAvailable;

#define MPS_POINTS 0
#define MPS_LINES 1
#define MPS_TRIANGLES 2
#define MPS_QUADS 3
#define MPS_POINTELEMENTS 4
#define MPS_POINTARRAYS 5
#define MPS_POINTELEMENTSLOCKED 6
#define MPS_POINTARRAYSLOCKED 7

static char *mpsMethodNames [] = {"MPS_POINTS", "MPS_LINES", "MPS_TRIANGLES", "MPS_QUADS", "MPS_POINTELEMENTS", "MPS_POINTARRAYS", "MPS_POINTELEMENTSLOCKED", "MPS_POINTARRAYSLOCKED"};
#define MPS_DRAWMETHODMAX (sizeof (mpsMethodNames) / sizeof (char *))

typedef void (*mpsDrawMethodType) (void);

typedef (*glLockArraysEXTType) (GLint first, GLsizei count);
typedef (*glUnlockArraysEXTType) (void);

static glLockArraysEXTType glLockArraysEXT;
static glUnlockArraysEXTType glUnlockArraysEXT;

int drawtype;
int lockArrays;
int drawElements;

static void mpsDrawPoints (void)
{
  int i;
  float *c;
  float *v;

  c = mpsColors [0];
  v = mpsVectors [0];

  if (mpsPointSize)
    glPointSize (mpsPointSize);

  glBegin (GL_POINTS);
	for (i = 0; i < mpsLength; i++)
  {
    glColor4fv (c);
    c += 4;
	  glVertex3fv (v);
    v += 3;
	}
	glEnd();
}


static void mpsDrawLines (void)
{
  int i;
  float x, y, z;

  if (mpsPointSize)
    glLineWidth (mpsPointSize);

  glBegin(GL_LINES);
	for (i = 0; i < mpsLength; i++)
  {
    x = mpsVectors [i][0];
    y = mpsVectors [i][1];
    z = mpsVectors [i][2];

    glColor4fv (mpsColors [i]);
	  glVertex3f(x, y, z);
    glColor4fv (mpsColors [i]);
	  glVertex3f(x+mpsPointSize*mpsDs [i][0],y+mpsPointSize*mpsDs [i][1],z+mpsPointSize*mpsDs [i][2]);
	}
	glEnd();
}

void mpsDrawTriangles (void)
{
  int i;
  float x, y, z;
  
  glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
  glDisable (GL_TEXTURE_2D);

	glBegin(GL_TRIANGLES);
	for (i=0; i< mpsLength; i++) {
    x = mpsVectors [i][0];
    y = mpsVectors [i][1];
    z = mpsVectors [i][2];

    glColor4fv (mpsColors [i]);
	  glVertex3f(x,y-(mpsPointSize/30),z);
	  glVertex3f(x, y+(mpsPointSize/30),z);
    glColor4fv (mpsColors [i]);
	  glVertex3f(x+mpsPointSize*mpsDs [i][0], y+mpsPointSize*mpsDs [i][1], z+mpsPointSize*mpsDs [i][2]);
	}
	glEnd();
}

void mpsDrawQuads (void)
{
  int i;
  float x, y, z;
  float dx, dy, dz;
  
  glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
  glDisable (GL_TEXTURE_2D);

	glBegin(GL_QUADS);
	for (i=0; i<mpsLength; i++) {
    x = mpsVectors [i][0];
    y = mpsVectors [i][1];
    z = mpsVectors [i][2];
    dx = mpsPointSize*mpsDs [i][0];
    dy = mpsPointSize*mpsDs [i][1];
    dz = mpsPointSize*mpsDs [i][2];

    glColor4fv (mpsColors [i]);
	  glVertex3f(x,y-(mpsPointSize/30),z);
	  glVertex3f(x, y+(mpsPointSize/30),z);
    glColor4fv (mpsColors [i]);
	  glVertex3f(x+dx, y+dy+(mpsPointSize/30), z+dz);
	  glVertex3f(x+dx, y+dy-(mpsPointSize/30), z+dz);
	}
	glEnd();
}

static void mpsDrawVectors (void)
{
  if (mpsPointSize)
    glPointSize(mpsPointSize);
 
  glEnableClientState (GL_VERTEX_ARRAY);
  glEnableClientState (GL_COLOR_ARRAY);
  glVertexPointer (3,GL_FLOAT,0, mpsVectors);
  glColorPointer (4,GL_FLOAT,0, mpsColors);
  if (lockArrays)
    if (glLockArraysEXT)
      glLockArraysEXT (0, mpsLength);

  if (drawElements)
    glDrawElements(GL_POINTS,mpsLength,GL_UNSIGNED_INT, mpsIndices);
  else
    glDrawArrays (GL_POINTS, 0, mpsLength);

  if (lockArrays)
    if (glUnlockArraysEXT)
      glUnlockArraysEXT ();

  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
}

void mpsDrawPointElementsLocked (void)
{
  drawElements = 1;
  lockArrays = 1;

  mpsDrawVectors ();
}

void mpsDrawPointElements (void)
{
  drawElements = 1;
  lockArrays = 0;

  mpsDrawVectors ();
}

void mpsDrawPointArraysLocked (void)
{
  drawElements = 0;
  lockArrays = 1;

  mpsDrawVectors ();
}

void mpsDrawPointArrays (void)
{
  drawElements = 0;
  lockArrays = 0;

  mpsDrawVectors ();
}

void mpsSetupPointSize (float size)
{
  if (size > 0)
  {
    if (size <= mpsMaxPointSize)
      mpsPointSize = size;
  }
}



/*
void drawtriangleelements (void)
{
  int indices [MAX_POINTS];
  static GLfloat colors[3*MAX_POINTS][4];
  static GLfloat triangles [9*MAX_POINTS][3];
  int i;
  int j;
  j = 0;
  for (i = 0; i < MAX_POINTS; i++)
  {
    indices [i] = i;
    memcpy (colors [j], mpsCol [pColors [i]], sizeof (GLfloat)*4);
    memcpy (colors [j+1], mpsCol [pColors [i]], sizeof (GLfloat)*4);
    memcpy (colors [j+2], mpsCol [pColors [i]], sizeof (GLfloat)*4);
    triangles [j] [0] = x;
    triangles [j] [1] = y;
    triangles [j] [2] = z;

    triangles [j+1] [0] = x;
    triangles [j+1] [1] = y+((mpsPointSize/30);
    triangles [j+1] [2] = z;

    triangles [j+2] [0] = x;
    triangles [j+2] [1] = y;
    triangles [j+2] [2] = z+((mpsPointSize/30);
    j += 3;
  }

  glEnableClientState (GL_VERTEX_ARRAY);
  glEnableClientState (GL_COLOR_ARRAY);
  glVertexPointer (3,GL_FLOAT,0, triangles);
  glColorPointer (4,GL_FLOAT,0, colors);
  if (lockArrays)
    if (glLockArraysEXT)
      glLockArraysEXT (0, 3*MAX_POINTS);

//  glDrawElements(GL_POINTS,MAX_POINTS,GL_UNSIGNED_INT, indices);
  glDrawArrays (GL_TRIANGLES, 0, 3*MAX_POINTS);

  if (lockArrays)
    if (glUnlockArraysEXT)
      glUnlockArraysEXT ();

  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
}
*/

void mpsInit (void)
{
  GLfloat ranges [10];  
  
#ifdef WIN32
  glLockArraysEXT = (glLockArraysEXTType) wglGetProcAddress("glLockArraysEXT");
  glUnlockArraysEXT = (glUnlockArraysEXTType)wglGetProcAddress("glUnlockArraysEXT");

  if (glLockArraysEXT && glUnlockArraysEXT)
     mpsLockArraysAvailable = 1;
  else
     mpsLockArraysAvailable = 0;
  #endif

  glGetFloatv (GL_POINT_SIZE_RANGE, ranges);
  mpsMaxPointSize = ranges [1];
}

void mpsSetupVector (float *vector)
{
  mpsVectors = (mpsVectorsType *) vector;
}

void mpsSetupD (float *d)
{
  mpsDs = (mpsDsType *) d;
}

void mpsSetupColors (float *colors)
{
  mpsColors = (mpsColorsType *) colors;
}

void mpsSetupLength (int length)
{
  int i;

  if (mpsLength != length)
  {
    if (!mpsIndices)
    {
      mpsIndices = malloc (length * sizeof (int) );
      mpsLength = 0;
    }
    else
      mpsIndices = realloc (mpsIndices, length * sizeof (int) );

    if (!mpsIndices)
      exit (1); // out of memory
    for (i = mpsLength; i < length; i++)
    {
      mpsIndices [i] = i;   
    }

    for (i = mpsLength; i < length; i++)
    {
      mpsIndices [i] = i;
    }

    mpsLength = length;
  }
}

void mpsSetupParameters (void)
{

}

mpsDrawMethodType mpsDrawMethod [MPS_DRAWMETHODMAX] = {mpsDrawPoints, mpsDrawLines, mpsDrawTriangles, mpsDrawQuads, mpsDrawPointElements, mpsDrawPointArrays, mpsDrawPointElementsLocked, mpsDrawPointArraysLocked};

void mpsDraw (int method)
{
  if (method > MPS_DRAWMETHODMAX)
    method = method % MPS_DRAWMETHODMAX;

  if (mpsVectors && mpsDs && mpsColors)
    mpsDrawMethod [method] ();
}

char *mpsGetMethodName (int method)
{
  if (method > MPS_DRAWMETHODMAX)
    method = method % MPS_DRAWMETHODMAX;

  return mpsMethodNames [method];
}
