
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>       /* for cos(), sin(), and sqrt() */
#include <time.h>
#include <GL/glut.h>

#include "mps.h"
#include "mglps.c"

/* Some <math.h> files do not define M_PI... */
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


#define CAMERA_DIST 30
#define CAMERA_ANGLE 20
#define BASE_VELOCITY 0.05
#define GRAVITY 0.005
#define LIFE_TIME_MAX 100

int maxPoints = 10000;

// time and points counters
int addMaxPoints = 100;
int numberPoints = 0;
int timer = 0;
int timerMax = 5;

// definition of the translations
GLfloat Xmove = -5.0;
GLfloat Ymove = 5.0;
GLfloat Zmove = 0;
// 10000 - 53 47 38 50 33 30 16.5 30
// definition of the angles
GLfloat XAngle = 30;
GLfloat YAngle = 240;
GLfloat ZAngle = 0;
// 100 000 - 5.8 5 3.9 5.15 3.33 1.7 3.3

int drawMethod;

int paused;

// definition of all the needed colours
//static GLfloat colorSet[][4] = {
  /* Shades of blue. */
/*
  { 0.0, 0.0, 1.0, 0.5 },
  { 0.0, 0.0, 1.0, 0.5 },
  { 0.0, 0.0, 1.0, 0.5 },
  { 0.0, 0.0, 5.0, 0.5 },
  { 0.0, 0.0, 5.0, 0.5 },
  { 0.0, 0.0, 5.0, 0.5 },
  { 0.0, 0.0, 1.0, 0.5 },
  { 0.0, 0.0, 1.0, 0.7 },
  { 0.0, 0.0, 1.0, 1.0 },
  { 0.0, 0.0, 5.0, 1.0 },
  { 0.0, 0.0, 5.0, 1.0 },
  { 0.0, 0.0, 1.0, 1.0 },
  { 2.0, 2.0, 1.0, 0.5 },
  { 1.0, 1.0, 1.0, 1.0 },
};
*/

/*
#define NUM_COLORS (sizeof(colorSet)/sizeof(colorSet[0]))

GLfloat black[] = {0.0, 0.0, 0.0};
GLfloat white[] = {1.0, 1.0, 1.0};
GLfloat red[] = {1.0, 0.0, 0.0};
GLfloat yellow[] = {1.0, 1.0, 0.0};
GLfloat green[] = {0.0, 1.0, 0.0};
GLfloat blue[] = {0.0, 0.0, 1.0};
GLfloat grey40[] = {0.6, 0.6, 0.6};
GLfloat grey10[] = {0.9, 0.9, 0.9};
GLfloat brown[] = {0.4, 0.1, 0.1};
*/

// definition of the structures needed to store
// information about the particles.

//GLfloat *pList [3];

int *lifeTime;

typedef GLfloat pListT [3];
typedef GLfloat pColorsT [4];
pListT *pList;
pListT *pDs;
pListT *pVelocity;
pColorsT *pColors;

// Function prototypes (needed for the forward referencing in main)
void draw(void);
void idle(void);
void key(unsigned char k, int x, int y);
void reshape(int width, int height);
void init(void);


// general changes to be made not to often (ie gravity, colors)
void updateChanges(void)
{
  int i;

  for (i=0; i<maxPoints; i++) {
//    if ((rand() % 10) > 6){
      //mpsColors[i] = rand() % NUM_COLORS;}
    // offsetX = ((float)(rand() % 10 -  2) / 1000);
    // offsetY = ((float)(rand() % 10 -  2) / 1000);
    // pVelocity[i][0] += offsetX;
    // pVelocity[i][1] += offsetY;
    pVelocity[i][1] -= GRAVITY;
    lifeTime[i] ++;
  }

}

// creates new particles(=points), computes all their items and stores them. 
void createPoints(void)
{

  float phi, teta, velocity;
  int i, newNumberPoints;

  
  newNumberPoints = numberPoints + (rand() % addMaxPoints);
  
  if (newNumberPoints > maxPoints)
    {
      numberPoints = 0;
      newNumberPoints = (rand() % addMaxPoints);
    }

  for (i=numberPoints; i<newNumberPoints; i++) { 
    pList[i][0] = Xmove;
    pList[i][1] = Ymove;
    pList[i][2] = Zmove;
    teta = ((float) (rand() % 25)) * (M_PI/180.0);
    phi = ((float) (rand() % 25)) * (M_PI/180.0);
    // velocity = BASE_VELOCITY;
    velocity = BASE_VELOCITY + ((float) (rand() % 30) / 1000) ;
    // velocity = (float)(rand() % 5) / 100;
    pVelocity[i][0] = velocity * cos(phi) * cos(teta);
    pVelocity[i][1] = velocity * sin(phi) * cos(teta);
    pVelocity[i][2] = velocity * sin(teta); 
    lifeTime[i] = 0;
  }

  numberPoints = newNumberPoints;
}

// changes to be made very often : each time we are going to redraw the scene.
void updatePoints(void)
{
  int i;

  for (i=0; i<maxPoints; i++) {
    if (lifeTime[i] < LIFE_TIME_MAX){
      pList[i][0] += pVelocity[i][0];
      if (pList[i][1] > 0.0){
	pList[i][1] += pVelocity[i][1];}
      pList[i][2] += pVelocity[i][2];
    }
    }
}



int count;

void initr (void)
{
        // Use a perspective projection for viewing:
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glFrustum(-2.0, 2.0, -2.0, 2.0, 6.0, 50.0);

        // Locate the camera down along the Z-axis
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glTranslatef(0.0, 0.0, -CAMERA_DIST);
        glRotatef(CAMERA_ANGLE, 1.0, 0.0, 0.0);
}

// The window size has changed so update the eye to viewport conversion
void reshape(int width, int height)
{
        glViewport(0, 0, (GLint)width, (GLint)height);


        initr ();
}


// redraws the whole scene.
void draw(void)
{
//  float fr;
//  GLUquadricObj *obj;

  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); 

//  initr ();

/*
  // default material properties:
        glMaterialfv(GL_FRONT, GL_SPECULAR, white);
        glMaterialf(GL_FRONT, GL_SHININESS, 10.0);
        glMaterialfv(GL_BACK, GL_DIFFUSE, black);
        glMaterialfv(GL_BACK, GL_SPECULAR, black);
        */

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(0.0, 0.0, 30.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

        glRotatef(XAngle, 1.0, 0.0, 0.0);
	glRotatef(YAngle, 0.0, 1.0, 0.0);
	glRotatef(ZAngle, 0.0, 0.0, 1.0);

  mpsSetupVector ((float *)pList);
  mpsSetupColors ((float *)pColors);
  mpsSetupD ((float *)pDs);
  mpsSetupLength (maxPoints);
  mpsDraw (drawMethod);
  
  mglpsInfo (drawMethod);

         /*
        glPushMatrix();         // save original frame

  obj = gluNewQuadric();

	glTranslatef(Xmove, Ymove, Zmove);

        // brown spherical water fountain
        // glRotatef(90, 1.0, 0.0, 0.0);
        // glTranslatef(-2.0, 5.0, 0.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, brown);
        gluSphere(obj, 1.0, 50, 50);

        glPopMatrix();          // restore original frame
        gluDeleteQuadric(obj);
        */
//  glFlush ();
  glutSwapBuffers ();
}


// Called when there is nothing else to do...
// This is how we normally animate i.e. redraw the window when
// the windowing system isn't up to anything else.
void idle(void)
{
  if (!paused)
  {
    timer ++;

    if (timer == timerMax)
      {
        updateChanges();
        createPoints();
        timer = 0;
      }
    updatePoints();      
  }
 
  draw();
}

// Handle keyboard input from the user.
void key(unsigned char k, int x, int y)
{
        switch (k) 
        {
        case 27:  /* Escape */
                exit(0);
                break;
        case 'x':       // rotate about x-axis
                XAngle += 5;
                break;
        case 'y':       // rotate about y-axis
                YAngle += 5;
                break;
        case 'z':       // rotate about z-axis
                ZAngle += 5;
                break;
	case 's':
                timerMax += 1;
                break;
        case 'S':
                timerMax -= 1;
                break;
/*
	case '+':
                addMaxPoints += 25;
                break;
	case '-':
                addMaxPoints -= 25;
                break;
*/
	case 'a':
                Xmove += 0.25;
                break;
	case 'A':
                Xmove -= 0.25;
                break;
	case 'b':
                Ymove += 0.25;
                break;
	case 'B':
                Ymove -= 0.25;
                break;
	case 'c':
                Zmove += 0.25;
                break;
	case 'C':
                Zmove -= 0.25;
                break;
  case 'p':
        paused = !paused;
        break;
  case '-':
    mpsSetupPointSize (mpsPointSize - 0.1);
    break;
  case '=':
    mpsSetupPointSize (mpsPointSize + 0.1);
    break;
  }

  if ((k >= '1') && (k <= '9'))
  {
    k = k - '1';
    if (k < MPS_DRAWMETHODMAX) 
      drawMethod = k; // change the current draw method
  }
}

float getrandcolor (void)
{
  return ((float) (rand () % 1000)) / 1000;
}

// Initialise the OpenGL state:
void init(void)
{
  int i;


  pList = (pListT *) malloc (sizeof (*pList) * maxPoints);
  pDs = (pListT *) malloc (sizeof (*pList) * maxPoints);
  pVelocity = (pListT *) malloc (sizeof (*pList) * maxPoints);
  pColors = (pColorsT *) malloc (sizeof (*pColors) * maxPoints);
  lifeTime = (int *) malloc (sizeof (int) * maxPoints);
  /*
        GLfloat light_pos[] = {5.0, 10.0, 50.0, 0.0};
        GLfloat ambient[] = {0.1, 0.1, 0.1, 0.0};

        glDisable(GL_DITHER);
//        glShadeModel(GL_SMOOTH);
        glDepthFunc(GL_LEQUAL);
        glEnable(GL_DEPTH_TEST);
        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);

        glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
  */

//  inittextures();
  for (i = 0; i < maxPoints; i++)
  {
    pDs [i] [0] = -0.1;
    pDs [i] [1] = 0.0;
    pDs [i] [2] = 0.0;
    pColors [i] [0] = 1;
    pColors [i] [1] = getrandcolor ();
    pColors [i] [2] = 1;
    pColors [i] [3] = 1;
  }
  mpsInit();
  mglpsInit ();
}

// Main function must pass any command line arguments to the GLUT
// system.  We also create and initialise all windows here.
main(int argc, char *argv[])
{
  int mpoints = 0;
  char title [999];

        if (argc >= 2)
        {
          sscanf (argv [1], "%i", &mpoints);
        }
        if (mpoints)
          maxPoints = mpoints;

        sprintf (title, "MPS Particle system - %i particles", maxPoints);

        glutInit(&argc, argv);
        glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
        glutCreateWindow(title);
        glutReshapeWindow(500, 500);

        init ();
        /*
        malloc (pList, sizeof (*plist) * maxPoints);
GLfloat pDs [maxPoints][3];
GLfloat pVelocity[maxPoints][3];
GLfloat pColors [maxPoints][4];
int lifeTime[maxPoints];*/

        // Install the message handlers
        glutReshapeFunc(reshape);
        glutKeyboardFunc(key);
        glutIdleFunc(idle);
        glutDisplayFunc(draw);

	createPoints();

        // Enter the main event loop.  This never terminates.
        glutMainLoop();

        return 0;             /* ANSI C requires main to return int. */
}



/*
  case '1': // 45 FPS
    drawproc = drawpoints;
    break;
  case '2': // 26 fps
    drawproc = drawtriangles;
    break;
  case '3': // 20 fps
    drawproc = drawquads;
    break;
  case '4': // 37 fps
    drawproc = drawpixelelements;
    lockArrays = 0;
    break;
  case '5': // 23 fps
    drawproc = drawpixelelements;
    lockArrays = 1;
    break;
  case '6': // 14.5 fps
    drawproc = drawtriangleelements;
    lockArrays = 0;
    break;
  case '7': // 8.7 fps
    drawproc = drawtriangleelements;
    lockArrays = 1;
	break;
  case '8': // 8.7 fps
    drawproc = drawlines;
    lockArrays = 1;
	break;
  */